"""
This module contains classes regarding the Amoeba potential and loading in a
TINKER-based parameter file.
"""
import logging
from ..utils.six.moves import range
from ..exceptions import TinkerError, TinkerWarning
from collections import OrderedDict
import re
import warnings

LOGGER = logging.getLogger(__name__)

class BookmarkedFile(object):
    """ Allows setting a bookmark and rewinding to that bookmark """

    def __init__(self, *args, **kwargs):
        self._stream = open(*args, **kwargs)

    def mark(self):
        self.bookmark = self.tell()

    def rewind(self):
        self.seek(self.bookmark)

    def __getattr__(self, attr):
        return getattr(self._stream, attr)

    def __del__(self):
        try:
            self._stream.close()
        except:
            pass

def _IS_INT(thing):
    try:
        int(thing)
        return True
    except ValueError:
        return False

def _IS_FLOAT(thing):
    try:
        float(thing)
        return True
    except ValueError:
        return False

#==============================================================================

class _ParamType(object):
    " All parameter types. This caches the list of parameters for easy access "
    TypeList = dict()
    _param_type = ''

    def __init__(self, *args, **kwargs):
        """ Instantiates the parameter type """
        raise NotImplementedError('virtual method')

    @classmethod
    def register(cls, obj, key):
        if key in cls.TypeList:
            warnmsg = 'Duplicate %s type found: %s' % (cls._param_type, key)
            if cls.TypeList[key] == obj:
                warnmsg += ' [same parameters]'
            else:
                warnmsg += ' [different parameters]'
            warnings.warn(warnmsg, TinkerWarning)
        cls.TypeList[key] = obj

    @classmethod
    def reset(cls):
        """
        Resets the cached list -- allows us to read in multiple parameter sets
        """
        cls.TypeList = dict()

    def __eq__(self, other):
        """ Make sure all attributes are the same """
        cls = type(self)
        if not isinstance(other, type(self)):
            raise TypeError('cannot compare type %s to type %s' %
                            (type(self), type(other)))
        for prop in dir(cls):
            if prop.startswith('_') or prop == 'TypeList': continue
            # Skip all callable attributes
            if hasattr(getattr(self, prop), '__call__'): continue
            if getattr(self, prop) != getattr(other, prop): return False

        return True

#==============================================================================

class _BondType(_ParamType):
    """ Bond parameter type """

    TypeList = dict()
    _param_type = 'bond'

    def __init__(self, idx1, idx2, k, req):
        idx1, idx2, = int(idx1), int(idx2)
        self.k, self.req = float(k), float(req)
        key = '%d-%d' % (min(idx1, idx2), max(idx1, idx2))
        self.register(self, key)

    def __repr__(self):
        return '<_BondType: k=%s; req=%s>' % (self.k, self.req)

#==============================================================================

def get_angle_type(typecode, *args, **kwargs):
    """ Factory that returns the appropriate angle type """
    if typecode in 'fF':
        return _FourierAngleType(*args, **kwargs)
    return _AngleType(*args, **kwargs)

class _AngleType(_ParamType):

    TypeList = dict()
    _param_type = 'angle'

    def __init__(self, idx1, idx2, idx3, k, theteq, theteq2=None, theteq3=None):
        idx1, idx2, idx3 = int(idx1), int(idx2), int(idx3)
        key = '%d-%d-%d' % (min(idx1, idx3), idx2, max(idx1, idx3))
        self.k, self.theteq = float(k), float(theteq)

        if theteq2 is not None:
            self.theteq2 = float(theteq2)
        else:
            self.theteq2 = None
        if theteq3 is not None:
            self.theteq3 = float(theteq3)
        else:
            self.theteq3 = None

        self.register(self, key)

    def __repr__(self):
        retval = "<_AngleType: k=%s; theteq=%s" % (self.k, self.theteq)
        if self.theteq2 is not None:
            retval += '; theteq2=%s' % self.theteq2
        if self.theteq3 is not None:
            retval += '; theteq3=%s' % self.theteq3
        return retval + '>'

class _FourierAngleType(_AngleType):

    def __init__(self, idx1, idx2, idx3, k, theteq, periodicity):
        idx1, idx2, idx3 = int(idx1), int(idx2), int(idx3)
        key = '%d-%d-%d' % (min(idx1, idx3), idx2, max(idx1, idx3))
        self.k = float(k)
        self.theteq = float(theteq)
        self.periodicity = float(periodicity)
        self.register(self, key)

    def __repr__(self):
        return '<_FourierAngleType: k=%s; theteq=%s; periodicity=%s>' % (
                    self.k, self.theteq, self.periodicity)

class _StretchBendType(_AngleType):
   
    TypeList = dict()
    _param_type = 'stretch-bend'

    def __init__(self, idx1, idx2, idx3, k1, k2):
        idx1, idx2, idx3 = int(idx1), int(idx2), int(idx3)
        key = '%d-%d-%d' % (min(idx1, idx3), idx2, max(idx1, idx3))
        self.k1, self.k2 = float(k1), float(k2)
        self.register(self, key)

    def __repr__(self):
        return '<_StretchBendType: k1=%s; k2=%s>' % (self.k1, self.k2)

class _UreyBradleyType(_AngleType):
   
    TypeList = dict()
    _param_type = 'urey-bradley'

    def __init__(self, idx1, idx2, idx3, k, req):
        idx1, idx2, idx3 = int(idx1), int(idx2), int(idx3)
        key = '%d-%d-%d' % (min(idx1, idx3), idx2, max(idx1, idx3))
        self.k, self.req = float(k), float(req)
        self.register(self, key)

    def __repr__(self):
        return '<_UreyBradleyType: k=%s; req=%s>' % (self.k, self.req)

#==============================================================================

class _OPBendType(_ParamType):

    TypeList = dict()
    _param_type = 'out-of-plane bending'

    def __init__(self, idx1, idx2, idx3, idx4, k):
        idx1, idx2, idx3, idx4 = int(idx1), int(idx2), int(idx3), int(idx4)
        self.k = float(k)
        key = '%d-%d-%d-%d' % (idx1, idx2, min(idx3, idx4), max(idx3, idx4))
        self.register(self, key)

    def __repr__(self):
        return '<_OPBendType: k=%s>' % self.k

#==============================================================================

class _DihedralType(_ParamType):

    TypeList = dict()
    _param_type = 'dihedral'

    def __init__(self, idx1, idx2, idx3, idx4, *args):
        idx1, idx2, idx3, idx4 = int(idx1), int(idx2), int(idx3), int(idx4)
        if idx2 < idx3 or (idx2 == idx3 and (idx1 < idx4 or idx1 == idx4)):
            key = '%d-%d-%d-%d' % (idx1, idx2, idx3, idx4)
        elif idx2 > idx3 or (idx2 == idx3 and idx1 > idx4):
            key = '%d-%d-%d-%d' % (idx4, idx3, idx2, idx1)
        self.k, self.phase, self.periodicity = [], [], []
        for i in range(len(args)//3):
            self.k.append(float(args[i*3]))
            self.phase.append(float(args[i*3+1]))
            self.periodicity.append(float(args[i*3+2]))
        self.register(self, key)

    def __repr__(self):
        return '<_DihedralType: k=%r; phase=%r; per=%r>' % (self.k, self.phase,
                self.periodicity)

class _PiTorsionType(_ParamType):

    TypeList = dict()
    _param_type = 'pi-torsion'

    def __init__(self, idx1, idx2, k):
        idx1, idx2 = int(idx1), int(idx2)
        key = '%d-%d' % (min(idx1, idx2), max(idx1, idx2))
        self.k = float(k)
        self.register(self, key)

    def __repr__(self):
        return '<_PiTorsionType: k=%s>' % self.k

class _TorsionTorsionType(_ParamType):

    TypeList = dict()
    _param_type = 'torsion-torsion'

    def __init__(self, indexes, nx, ny):
        indexes = (int(i) for i in indexes)
        key = '%d-%d-%d-%d-%d' % tuple(indexes)
        self.nx, self.ny = int(nx), int(ny)
        self.potential_grid = OrderedDict()
        self.register(self, key)

    def add_point(self, x, y, potential):
        self.potential_grid[(float(x), float(y))] = float(potential)

    def __repr__(self):
        return '<_TorsionTorsion: %d x %d potential grid>' % (self.nx, self.ny)

#==============================================================================

class _MultipoleType(_ParamType):

    TypeList = dict()
    _param_type = 'multipole'

    def __init__(self, indexes, p1):
        indexes = (str(i) for i in indexes)
        key = '-'.join(indexes)
        self.potential_terms = [float(p1)]
        self.register(self, key)

    def __repr__(self):
        return '<_MultipoleType: terms=%r>' % self.potential_terms

    def add_terms(self, terms):
        for term in terms:
            self.potential_terms.append(float(term))

#==============================================================================

def get_atom_type(index, atomic_number, mass, valence):
    """
    Factory for getting an _AtomType, but making sure that only one instance of
    a particular type is created
    """
    index = int(index)
    try:
        return _AtomType.TypeList[index]
    except KeyError:
        return _AtomType(index, atomic_number, mass, valence)

class _AtomType(object):
    """ An atom type """
    TypeList = dict() # All cached types -- ensures atom types are unique

    def __init__(self, index, atomic_number, mass, valence):
        self.index = int(index)
        self.atomic_number = int(atomic_number)
        self.mass = float(mass)
        self.valence = int(valence)
        _AtomType.TypeList[index] = self # cache this type

    @staticmethod
    def set_vdw_params(index, size, epsilon, reduction=None):
        inst = _AtomType.TypeList[int(index)]
        inst.size = float(size)
        inst.epsilon = float(epsilon)
        if reduction is not None:
            inst.reduction = float(reduction)
        else:
            inst.reduction = None

    def __repr__(self):
        retval = '<_AtomType: idx=%d; elem=%d; mass=%s; val=%d' % (
                       self.index, self.atomic_number, self.mass, self.valence)
        if hasattr(self, 'size'):
            retval += '; size=%s; eps=%s; red=%s' % (self.size, self.epsilon,
                self.reduction)
        return retval + '>'

    @classmethod
    def reset(cls):
        cls.TypeList = dict()

class _Atom(object):
    """ An atom in a parameter set """
    AtomList=dict()
    def __init__(self, index,typeindex, name, descrip, atomic_number, mass, val):
        self.name = name
        self.description = descrip
        self.type = get_atom_type(typeindex, atomic_number, mass, val)
        _Atom.AtomList[index] = self # cache this type

    # Allow _Atom instances to access (but not modify) type properties
    def _typeindex(self): return self.type.index
    def _atomic_number(self): return self.type.atomic_number
    def _element(self): return self.type.atomic_number
    def _valence(self): return self.type.valence
    def _size(self): return self.type.size
    def _epsilon(self): return self.type.epsilon
    def _reduction(self): return self.type.reduction
    def _blocked(self): raise NotImplementedError('Cannot set this attribute')
    # Now set the above as properties
    typeindex = property(fget=_typeindex, fset=_blocked)
    atomic_number = property(fget=_atomic_number, fset=_blocked)
    element = property(fget=_element, fset=_blocked)
    valence = property(fget=_valence, fset=_blocked)
    size = property(fget=_size, fset=_blocked)
    epsilon = property(fget=_epsilon, fset=_blocked)
    reduction = property(fget=_reduction, fset=_blocked)

    def set_polarizability(self, polarizability, thole, connected_types):
        self.polarizability = float(polarizability)
        self.thole = float(thole)
        self.connected_types = [int(i) for i in connected_types]

    def __repr__(self):
        retstr = '<_Atom "%s": name=%s; type=%d' % (
                            self.description, self.name, self.typeindex)
        if hasattr(self, 'polarizability'):
            retstr += '; dipole pol=%s; thole=%s; connected atoms=%r' % (
                        self.polarizability, self.thole, self.connected_types)
        return retstr+'>'

    @classmethod
    def reset(cls):
        cls.TypeList = dict()

#==============================================================================

def reset():
    """
    Resets all of the TypeList instances (without destroying the data inside
    them) so we can load multiple parameter sets
    """
    _BondType.reset()
    _AngleType.reset()
    _FourierAngleType.reset()
    _StretchBendType.reset()
    _UreyBradleyType.reset()
    _OPBendType.reset()
    _DihedralType.reset()
    _PiTorsionType.reset()
    _TorsionTorsionType.reset()
    _MultipoleType.reset()
    _AtomType.reset()

#==============================================================================

class AmoebaParameterSet(object):
    """
    Contains all of the parameters found in an Amoeba parameter file from TINKER
    """
    atomre = re.compile(r'atom *(\d+) *(\d+) *([A-Za-z\-\+\*0-9]+) *"(.+)" *'
                        r'(\d+) *(\d+\.\d+) *(\d+)', re.I)
    anglere = re.compile(r'angle([ 345fF])')

    def __init__(self, fname=None):
        self.atoms = dict()
        self.atom_types = _AtomType.TypeList # For easy access
        self.bonds = _BondType.TypeList
        self.angles = _AngleType.TypeList # includes FourierAngleTypes
        self.stretch_bends = _StretchBendType.TypeList
        self.urey_bradleys = _UreyBradleyType.TypeList
        self.opbends = _OPBendType.TypeList
        self.dihedrals = _DihedralType.TypeList
        self.torsion_torsions = _TorsionTorsionType.TypeList
        self.multipoles = _MultipoleType.TypeList
        if fname is not None:
            self.load_parameter_file(fname)

    def load_parameter_file(self, fname):
        """
        Parses a parameter file and loads all of the parameters found into data
        structures.
        """
        self.attributes = dict()

        # First load the attributes from the header
        f = BookmarkedFile(fname, 'r')
        done_with_attributes = False
        line = f.readline().replace('\t', ' ')
        while line and not done_with_attributes:
            if 'Literature References' in line:
                done_with_attributes = True
                break
            line = (line + '#').strip()
            line = line[:line.index('#')].strip()
            if not line:
                line = f.readline().replace('\t', ' ')
                continue
            words = line.split()
            if len(words) != 2: continue
            # Now extract the property
            if _IS_INT(words[1]):
                self.attributes[words[0].lower()] = int(words[1])
            elif _IS_FLOAT(words[1]):
                self.attributes[words[0].lower()] = float(words[1])
            else:
                self.attributes[words[0].lower()] = words[1]
            line = f.readline().replace('\t', ' ')
        if not done_with_attributes:
            raise TinkerError('Could not find force field attributes.')
        # Now get the atom types
        while line.lstrip()[:5].lower() != 'atom ':
            line = f.readline().replace('\t', ' ')
        # Now loop through all atoms
        while line.lstrip()[:5].lower() == 'atom ':
            rematch = self.atomre.match(line)
            num, typenum, name, descrip, anum, mass, val = rematch.groups()
            self.atoms[int(num)] = _Atom(int(num),typenum, name, descrip,
                                                 anum, mass, val)
            line = f.readline().replace('\t', ' ')
        # Now parse out the van der waals terms
        while line.lstrip()[:4].lower() != 'vdw ':
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:4].lower() == 'vdw ':
            _AtomType.set_vdw_params(*line.split()[1:])
            line = f.readline().replace('\t', ' ')
        # Now parse out the bonds
        while line.lstrip()[:5].lower() != 'bond ':
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:5].lower() == 'bond ':
            _BondType(*line.split()[1:])
            line = f.readline().replace('\t', ' ')
        # Now parse out the angles. Handle iring and Fourier terms
        rematch = self.anglere.match(line)
        while not rematch:
            line = f.readline().replace('\t', ' ')
            rematch = self.anglere.match(line)
        while rematch:
            try:
                get_angle_type(rematch.groups()[0], *line.split()[1:])
            except TypeError:
                LOGGER.debug('%s, %s', repr(rematch.groups()[0]), line.split()[1:])
                raise
            line = f.readline().replace('\t', ' ')
            rematch = self.anglere.match(line)
        # Now parse out the stretch-bend parameters. From here on out, some of
        # the terms may not exist in all versions of the force field, so
        # protect for EOF and make sure we rewind to avoid missing any terms.
        f.mark()
        while line.lstrip()[:7].lower() != 'strbnd ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:7].lower() == 'strbnd ' and line:
            f.mark()
            _StretchBendType(*line.split()[1:])
            line = f.readline().replace('\t', ' ')
        # Get the Urey-Bradley term(s)
        f.rewind(); line = f.readline().replace('\t', ' ')
        while line.lstrip()[:9].lower() != 'ureybrad ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:9].lower() == 'ureybrad ' and line:
            f.mark()
            _UreyBradleyType(*line.split()[1:])
            line = f.readline().replace('\t', ' ')
        # Get the out-of-plane bending
        f.rewind(); line = f.readline().replace('\t', ' ')
        while line.lstrip()[:7].lower() != 'opbend ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:7].lower() == 'opbend ' and line:
            f.mark()
            _OPBendType(*line.split()[1:])
            line = f.readline().replace('\t', ' ')
        # Get the torsion parameters
        f.rewind(); line = f.readline().replace('\t', ' ')
        while line.lstrip()[:8].lower() != 'torsion ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:8].lower() == 'torsion ' and line:
            f.mark()
            _DihedralType(*line.split()[1:])
            line = f.readline().replace('\t', ' ')
        # Get the pitorsions
        f.rewind(); line = f.readline().replace('\t', ' ')
        while line.lstrip()[:7] != 'pitors ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:7] == 'pitors ' and line:
            f.mark()
            _PiTorsionType(*line.split()[1:])
            line = f.readline().replace('\t', ' ')
        # Get the coupled torsions
        f.rewind(); line = f.readline().replace('\t', ' ')
        while line.lstrip()[:8] != 'tortors ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:8] == 'tortors ' and line:
            words = line.split()
            tortor = _TorsionTorsionType(words[1:6], words[6], words[7])
            line = f.readline().replace('\t', ' ')
            while line.strip():
                tortor.add_point(*line.split())
                line = f.readline().replace('\t', ' ')
            line = f.readline().replace('\t', ' ')
            f.mark()
        # Get the multipole terms
        f.rewind(); line = f.readline().replace('\t', ' ')
        while line.lstrip()[:10].lower() != 'multipole ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:10].lower() == 'multipole ' and line:
            words = line.split()
            multipole = _MultipoleType(words[1:-1], words[-1])
            multipole.add_terms(f.readline().split())
            multipole.add_terms(f.readline().split())
            multipole.add_terms(f.readline().split())
            multipole.add_terms(f.readline().split())
            line = f.readline().replace('\t', ' ')
            f.mark()
        # Get the dipole polarizabilities
        f.rewind(); line = f.readline().replace('\t', ' ')
        while line.lstrip()[:9] != 'polarize ' and line:
            line = f.readline().replace('\t', ' ')
        while line.lstrip()[:9] == 'polarize ' and line:
            words = line.split()
            index = int(words[1])
            try:
                self.atoms[index].set_polarizability(
                            words[2], words[3], words[4:]
                )
            except IndexError:
                self.atoms[index].set_polarizability(words[2], words[3], [])
            line = f.readline().replace('\t', ' ')
        f.close()
        # Now clean up so we can load another parameter set
        reset()
        return



"""
This module contains parsers for mol2-format files (with support for the mol3
extension described at http://q4md-forcefieldtools.org/Tutorial/leap-mol3.php
"""
from __future__ import print_function, division, absolute_import

import copy
from parmed.exceptions import Mol2Error, ParameterWarning
from parmed.formats.registry import FileFormatType
from parmed.modeller.residue import ResidueTemplate, ResidueTemplateContainer
from parmed.periodic_table import element_by_name, AtomicNum
from parmed.residue import AminoAcidResidue, RNAResidue, DNAResidue
from parmed.structure import Structure
from parmed.topologyobjects import Atom, Bond
from parmed.utils.io import genopen
from parmed.utils.six import add_metaclass, string_types
import warnings

@add_metaclass(FileFormatType)
class Mol2File(object):
    """ Class to read and write TRIPOS Mol2 files """

    BOND_ORDER_MAP = dict(ar=1.5, am=1.25)
    REVERSE_BOND_ORDER_MAP = {1.25 : 'am', 1.5 : 'ar'}

    #===================================================

    @staticmethod
    def id_format(filename):
        """ Identify the file as a Mol2 (or Mol3) file format or not

        Parameters
        ----------
        filename : str
            Name of the file to test whether or not it is a mol2 file

        Returns
        -------
        is_fmt : bool
            True if it is a mol2 (or mol3) file, False otherwise
        """
        f = genopen(filename, 'r')
        try:
            for line in f:
                if line.startswith('#'): continue
                if not line.strip(): continue
                return line.startswith('@<TRIPOS>')
            return False
        finally:
            f.close()

    #===================================================

    @staticmethod
    def parse(filename, structure=False):
        """ Parses a mol2 file (or mol3) file

        Parameters
        ----------
        filename : str or file-like
            Name of the file to parse or file-like object to parse from
        structure : bool, optional
            If True, the return value is a :class:`Structure` instance. If
            False, it is either a :class:`ResidueTemplate` or
            :class:`ResidueTemplateContainter` instance, depending on whether
            there is one or more than one residue defined in it. Default is
            False

        Returns
        -------
        molecule : :class:`Structure`, :class:`ResidueTemplate`, or
                   :class:`ResidueTemplateContainer`
            The molecule defined by this mol2 file

        Raises
        ------
        Mol2Error
            If the file format is not recognized or non-numeric values are
            present where integers or floating point numbers are expected. Also
            raises Mol2Error if you try to parse a mol2 file that has multiple
            @<MOLECULE> entries with ``structure=True``.
        """
        if isinstance(filename, string_types):
            f = genopen(filename, 'r')
            own_handle = True
        else:
            f = filename
            own_handle = False
        rescont = ResidueTemplateContainer()
        struct = Structure()
        restemp = ResidueTemplate()
        mol_info = []
        multires_structure = False
        try:
            section = None
            last_residue = None
            headtail = 'head'
            molecule_number = 0
            for line in f:
                if line.startswith('#'): continue
                if not line.strip() and section is None: continue
                if line.startswith('@<TRIPOS>'):
                    section = line[9:].strip()
                    if section == 'MOLECULE' and (restemp.atoms or rescont):
                        if structure:
                            raise Mol2Error('Cannot convert MOL2 with multiple '
                                            '@<MOLECULE>s to a Structure')
                        # Set the residue name from the MOL2 title if the
                        # molecule had only 1 residue and it was given a name in
                        # the title
                        if not multires_structure and mol_info[0]:
                            restemp.name = mol_info[0]
                        multires_structure = False
                        rescont.append(restemp)
                        restemp = ResidueTemplate()
                        struct = Structure()
                        last_residue = None
                        molecule_number += 1
                        mol_info = []
                    continue
                if section is None:
                    raise Mol2Error('Bad mol2 file format')
                if section == 'MOLECULE':
                    # Section formatted as follows:
                    #   mol_name
                    #   num_atoms [num_bonds [num_substr [num_feat [num_sets]]]]
                    #   mol_type
                    #   charge_type
                    #   [status_bits]
                    #   [mol_comment]
                    # TODO: Do something with the name.
                    if len(mol_info) == 0:
                        mol_info.append(line.strip())
                    elif len(mol_info) == 1:
                        mol_info.append([int(x) for x in line.split()])
                    elif len(mol_info) == 2:
                        mol_info.append(line.strip())
                    elif len(mol_info) == 3:
                        mol_info.append(line.strip())
                    # Ignore the rest
                    continue
                if section == 'ATOM':
                    # Section formatted as follows:
                    #   atom_id -- serial number of atom
                    #   atom_name -- name of the atom
                    #   x -- X-coordinate of the atom
                    #   y -- Y-coordinate of the atom
                    #   z -- Z-coordinate of the atom
                    #   atom_type -- type of the atom
                    #   subst_id -- Residue serial number
                    #   subst_name -- Residue name
                    #   charge -- partial atomic charge
                    #   status_bit -- ignored
                    words = line.split()
                    id = int(words[0])
                    name = words[1]
                    x = float(words[2])
                    y = float(words[3])
                    z = float(words[4])
                    typ = words[5]
                    try:
                        resid = int(words[6])
                    except IndexError:
                        resid = 0
                    try:
                        resname = words[7]
                    except IndexError:
                        resname = 'UNK'
                    if 'NO_CHARGES' not in mol_info:
                        try:
                            charge = float(words[8])
                        except IndexError:
                            charge = 0
                    else:
                        charge = 0
                    if last_residue is None:
                        last_residue = (resid, resname)
                        restemp.name = resname
                    atom = Atom(name=name, type=typ, number=id, charge=charge)
                    atom.xx, atom.xy, atom.xz = x, y, z
                    struct.add_atom(atom, resname, resid)
                    if last_residue != (resid, resname):
                        rescont.append(restemp)
                        restemp = ResidueTemplate()
                        restemp.name = resname
                        last_residue = (resid, resname)
                        multires_structure = True
                    try:
                        restemp.add_atom(copy.copy(atom))
                    except ValueError:
                        # Allow mol2 files being parsed as a Structure to have
                        # duplicate atom names
                        if not structure:
                            raise
                    continue
                if section == 'BOND':
                    # Section formatted as follows:
                    #   bond_id -- serial number of bond (ignored)
                    #   origin_atom_id -- serial number of first atom in bond
                    #   target_atom_id -- serial number of other atom in bond
                    #   bond_type -- string describing bond type
                    #   status_bits -- ignored
                    words = line.split()
                    int(words[0]) # Bond serial number... redundant and ignored
                    a1 = int(words[1])
                    a2 = int(words[2])
                    try:
                        order = words[3]
                    except IndexError:
                        order = 1.0
                    if order in Mol2File.BOND_ORDER_MAP:
                        order = Mol2File.BOND_ORDER_MAP[order]
                    else:
                        try:
                            order = float(order)
                        except ValueError:
                            warnings.warn('Mol2 bond order not recognized: %s' %
                                          order, ParameterWarning)
                            order = 1.0
                    atom1 = struct.atoms.find_original_index(a1)
                    atom2 = struct.atoms.find_original_index(a2)
                    struct.bonds.append(Bond(atom1, atom2, order=order))
                    # Now add it to our residue container
                    # See if it's a head/tail connection
                    if atom1.residue is not atom2.residue:
                        if atom1.residue.idx == len(rescont):
                            res1 = restemp
                        elif atom1.residue.idx < len(rescont):
                            res1 = rescont[atom1.residue.idx]
                        assert atom.residue.idx <= len(rescont), 'Bad bond!'
                        if atom2.residue.idx == len(rescont):
                            res2 = restemp
                        elif atom2.residue.idx < len(rescont):
                            res2 = rescont[atom2.residue.idx]
                        assert atom.residue.idx <= len(rescont), 'Bad bond!'
                        assert res1 is not res2, 'BAD identical residues'
                        idx1 = atom1.idx - atom1.residue[0].idx
                        idx2 = atom2.idx - atom2.residue[0].idx
                        if atom1.residue.idx < atom2.residue.idx:
                            res1.tail = res1[idx1]
                            res2.head = res2[idx2]
                        else:
                            res1.head = res1[idx1]
                            res2.tail = res2[idx2]
                    elif not multires_structure:
                        if not structure:
                            restemp.add_bond(a1-1, a2-1, order)
                    else:
                        # Same residue, add the bond
                        offset = atom1.residue[0].idx
                        if atom1.residue.idx == len(rescont):
                            res = restemp
                        else:
                            res = rescont[atom1.residue.idx]
                        res.add_bond(atom1.idx-offset, atom2.idx-offset, order)
                    continue
                if section == 'CRYSIN':
                    # Section formatted as follows:
                    #   a -- length of first unit cell vector
                    #   b -- length of second unit cell vector
                    #   c -- length of third unit cell vector
                    #   alpha -- angle b/w b and c
                    #   beta -- angle b/w a and c
                    #   gamma -- angle b/w a and b
                    #   space group -- number of space group (ignored)
                    #   space group setting -- ignored
                    words = line.split()
                    box = [float(w) for w in words[:6]]
                    if len(box) != 6:
                        raise ValueError('%d box dimensions found; needed 6' %
                                         len(box))
                    struct.box = copy.copy(box)
                    rescont.box = copy.copy(box)
                    continue
                if section == 'SUBSTRUCTURE':
                    # Section formatted as follows:
                    #   subst_id -- residue number
                    #   subst_name -- residue name
                    #   root_atom -- first atom of residue
                    #   subst_type -- ignored (usually 'RESIDUE')
                    #   dict_type -- type of substructure (ignored)
                    #   chain -- chain ID of residue
                    #   sub_type -- type of the chain
                    #   inter_bonds -- # of inter-substructure bonds
                    #   status -- ignored
                    #   comment -- ignored
                    words = line.split()
                    if not words: continue
                    id = int(words[0])
                    resname = words[1]
                    try:
                        chain = words[5]
                    except IndexError:
                        chain = ''
                    # Set the chain ID
                    for res in struct.residues:
                        if res.number == id and res.name == resname:
                            res.chain = chain
                    continue
                # MOL3 sections
                if section == 'HEADTAIL':
                    atname, residx = line.split()
                    residx = int(residx)
                    if residx in (0, 1) or residx - 1 == len(rescont):
                        res = restemp
                    elif residx - 1 < len(rescont):
                        res = rescont[residx-1]
                    else:
                        raise Mol2Error('Residue out of range in head/tail')
                    for atom in res:
                        if atom.name == atname:
                            if headtail == 'head':
                                res.head = atom
                                headtail = 'tail'
                            else:
                                res.tail = atom
                                headtail = 'head'
                            break
                    else:
                        if headtail == 'head':
                            headtail = 'tail'
                        else:
                            headtail = 'head'
                    continue
                if section == 'RESIDUECONNECT':
                    words = line.split()
                    residx = int(words[0])
                    if residx - 1 == len(rescont):
                        res = restemp
                    elif residx - 1 < len(rescont):
                        res = rescont[residx-1]
                    else:
                        raise Mol2Error('Residue out of range in '
                                        'residueconnect')
                    for a in words[3:]:
                        if a == '0': continue
                        for atom in res:
                            if atom.name == a:
                                res.connections.append(atom)
                                break
                        else:
                            raise Mol2Error('Residue connection atom %s not '
                                            'found in residue %d' % (a, residx))
            if structure:
                for atom in struct.atoms:
                    anum = _guess_atomic_number(atom.name, restemp)
                    if anum == 0:
                        anum = _guess_atomic_number(atom.type, restemp)
                    atom.atomic_number = anum
                return struct
            elif len(rescont) > 0:
                if not multires_structure and mol_info[0]:
                    restemp.name = mol_info[0]
                rescont.append(restemp)
                for res in rescont:
                    for atom in res.atoms:
                        anum = _guess_atomic_number(atom.name, restemp)
                        if anum == 0:
                            anum = _guess_atomic_number(atom.type, restemp)
                        atom.atomic_number = anum
                return rescont
            else:
                for atom in restemp.atoms:
                    anum = _guess_atomic_number(atom.name, restemp)
                    if anum == 0:
                        anum = _guess_atomic_number(atom.type, restemp)
                    atom.atomic_number = anum
                return restemp
        except ValueError as e:
            raise Mol2Error('String conversion trouble: %s' % e)
        finally:
            if own_handle: f.close()

    #===================================================

    @staticmethod
    def write(struct, dest, mol3=False, split=False, compress_whitespace=False):
        """ Writes a mol2 file from a structure or residue template

        Parameters
        ----------
        struct : :class:`Structure` or :class:`ResidueTemplate` or
                 :class:`ResidueTemplateContainer`
            The input structure to write the mol2 file from
        dest : str or file-like obj
            Name of the file to write or open file handle to write to
        mol3 : bool, optional
            If True and ``struct`` is a ResidueTemplate or container, write
            HEAD/TAIL sections. Default is False
        split : bool, optional
            If True and ``struct`` is a ResidueTemplateContainer or a Structure
            with multiple residues, each residue is printed in a separate
            @<MOLECULE> section that appear sequentially in the output file
        compress_whitespace : bool, optional
            If True, seprate fields on one line with a single space instead of
            aligning them with whitespace. This is useful for parsers that
            truncate lines at 80 characters (e.g., some versions of OpenEye).
            However, it will not look as "neat" upon visual inspection in a text
            editor. Default is False.
        """
        own_handle = False
        if not hasattr(dest, 'write'):
            own_handle = True
            dest = genopen(dest, 'w')
        if split:
            # Write sequentially if it is a multi-residue container or Structure
            if isinstance(struct, ResidueTemplateContainer):
                try:
                    for res in struct:
                        Mol2File.write(res, dest, mol3,
                                       compress_whitespace=compress_whitespace)
                finally:
                    if own_handle: dest.close()
                return
            elif isinstance(struct, Structure) and len(struct.residues) > 1:
                try:
                    for res in ResidueTemplateContainer.from_structure(struct):
                        Mol2File.write(res, dest, mol3,
                                       compress_whitespace=compress_whitespace)
                finally:
                    if own_handle: dest.close()
                return
        try:
            if isinstance(struct, ResidueTemplateContainer):
                natom = sum([len(c) for c in struct])
                # To find the number of bonds, we need to total number of bonds
                # + the number of bonds that would be formed by "stitching"
                # together residues via their head and tail
                bonds = []
                charges = []
                bases = [1 for res in struct]
                for i, res in enumerate(struct):
                    if i < len(struct) - 1:
                        bases[i+1] = bases[i] + len(res)
                for i, res in enumerate(struct):
                    for bond in res.bonds:
                        bonds.append((bond.atom1.idx+bases[i],
                                      bond.atom2.idx+bases[i],
                                      bond.order))
                    if i < len(struct)-1 and (res.tail is not None and
                            struct[i+1].head is not None):
                        bonds.append((res.tail.idx+bases[i],
                                      struct[i+1].head.idx+bases[i+1],
                                      bond.order))
                    charges.extend([a.charge for a in res])
                residues = struct
                name = struct.name or struct[0].name
            else:
                natom = len(struct.atoms)
                bonds = [(b.atom1.idx+1, b.atom2.idx+1, b.order)
                            for b in struct.bonds]
                if isinstance(struct, ResidueTemplate):
                    residues = [struct]
                    name = struct.name
                else:
                    residues = struct.residues
                    name = struct.residues[0].name
                charges = [a.charge for a in struct.atoms]
            dest.write('@<TRIPOS>MOLECULE\n')
            dest.write('%s\n' % name)
            dest.write('%d %d %d 0 1\n' % (natom, len(bonds), len(residues)))
            if len(residues) == 1:
                dest.write('SMALL\n')
            else:
                for residue in residues:
                    if AminoAcidResidue.has(residue.name):
                        dest.write('PROTEIN\n')
                        break
                    if (RNAResidue.has(residue.name) or
                            DNAResidue.has(residue.name)):
                        dest.write('NUCLEIC\n')
                        break
                else:
                    dest.write('BIOPOLYMER\n')
            if not any(charges):
                dest.write('NO_CHARGES\n')
                printchg = False
            else:
                dest.write('USER_CHARGES\n')
                printchg = True
            # See if we want to print box info
            if hasattr(struct, 'box') and struct.box is not None:
                box = struct.box
                dest.write('@<TRIPOS>CRYSIN\n')
                if compress_whitespace:
                    fmt = '%.4f %.4f %.4f %.4f %.4f %.4f 1 1\n'
                else:
                    fmt = '%10.4f %10.4f %10.4f %10.4f %10.4f %10.4f  1  1\n'
                dest.write(fmt % tuple(box))
            # Now do ATOM section
            dest.write('@<TRIPOS>ATOM\n')
            j = 1
            for i, res in enumerate(residues):
                for atom in res:
                    try:
                        x = atom.xx
                    except AttributeError:
                        x = 0
                    try:
                        y = atom.xy
                    except AttributeError:
                        y = 0
                    try:
                        z = atom.xz
                    except AttributeError:
                        z = 0
                    if compress_whitespace:
                        fmt = '%d %s %.4f %.4f %.4f %s %d %s'
                    else:
                        fmt = '%8d %-8s %10.4f %10.4f %10.4f %-8s %6d %-8s'
                    dest.write(fmt % (j, atom.name, x, y, z,
                               atom.type.strip() or atom.name, i+1, res.name))
                    if printchg:
                        if compress_whitespace:
                            fmt = ' %.6f\n'
                        else:
                            fmt = ' %10.6f\n'
                        dest.write(fmt % atom.charge)
                    else:
                        dest.write('\n')
                    j += 1
            dest.write('@<TRIPOS>BOND\n')
            for i, bond in enumerate(bonds):
                if bond[2] in Mol2File.REVERSE_BOND_ORDER_MAP:
                    order = Mol2File.REVERSE_BOND_ORDER_MAP[bond[2]]
                else:
                    order = int(bond[2])
                if compress_whitespace:
                    fmt = '%d %d %d %s\n'
                else:
                    fmt = '%8d %8d %8d %s\n'
                dest.write(fmt % (i+1, bond[0], bond[1], order))
            dest.write('@<TRIPOS>SUBSTRUCTURE\n')
            first_atom = 0
            for i, res in enumerate(residues):
                if not hasattr(res, 'chain') or not res.chain:
                    chain = '****'
                else:
                    chain = res.chain
                intresbonds = 0
                if isinstance(res, ResidueTemplate):
                    if i != len(residues)-1 and (res.tail is not None and
                            residues[i+1].head is not None):
                        intresbonds += 1
                    if i != 0 and (res.head is not None and residues[i-1].tail
                            is not None):
                        intresbonds += 1
                else:
                    for atom in res:
                        for a2 in atom.bond_partners:
                            if a2.residue is not res:
                                intresbonds += 1
                if compress_whitespace:
                    fmt = '%d %s %d RESIDUE %d %s ROOT %d\n'
                else:
                    fmt = '%8d %-8s %8d RESIDUE %4d %-4s ROOT %6d\n'
                dest.write(fmt % (i+1, res.name, first_atom+1, 0, chain[:4],
                           intresbonds))
                first_atom += len(res)
            if mol3:
                dest.write('@<TRIPOS>HEADTAIL\n')
                for i, res in enumerate(residues):
                    if isinstance(res, ResidueTemplate):
                        if res.head is not None:
                            dest.write('%s %d\n' % (res.head.name, i+1))
                        else:
                            dest.write('0 0\n')
                        if res.tail is not None:
                            dest.write('%s %d\n' % (res.tail.name, i+1))
                        else:
                            dest.write('0 0\n')
                    else:
                        head = tail = None
                        for atom in res:
                            for a2 in atom.bond_partners:
                                if a2.residue.idx == res.idx - 1:
                                    head = atom
                                if a2.residue.idx == res.idx + 1:
                                    tail = atom
                        if head is not None:
                            dest.write('%s %d\n' % (head.name, i+1))
                        else:
                            dest.write('0 0\n')
                        if tail is not None:
                            dest.write('%s %d\n' % (tail.name, i+1))
                        else:
                            dest.write('0 0\n')
                dest.write('@<TRIPOS>RESIDUECONNECT\n')
                for i, res in enumerate(residues):
                    if isinstance(res, ResidueTemplate):
                        con = [res.head, res.tail, None, None, None, None]
                        for i, a in enumerate(res.connections):
                            con[i+2] = a
                    else:
                        con = [None, None, None, None, None, None]
                        ncon = 2
                        for atom in res:
                            for a2 in atom.bond_partners:
                                if a2.residue.idx == res.idx - 1:
                                    con[0] = atom
                                elif a2.residue.idx == res.idx + 1:
                                    con[1] = atom
                                elif a2.residue.idx != res.idx:
                                    con[ncon] = atom
                                    ncon += 1
                    dest.write('%d' % (i+1))
                    for a in con:
                        if a is not None:
                            dest.write(' %s' % a.name)
                        else:
                            dest.write(' 0')
                    dest.write('\n')
        finally:
            if own_handle: dest.close()

def _guess_atomic_number(name, residue=None):
    """ Guesses the atomic number """
    # Special-case single-atom residues, which are almost always ions
    name = ''.join(c for c in name if c.isalpha())
    if residue is None or len(residue.atoms) == 1:
        if len(name) > 1:
            try:
                return AtomicNum[name[0].upper() + name[1].lower()]
            except KeyError:
                return AtomicNum[element_by_name(name)]
    return AtomicNum[element_by_name(name)]
